/* gpl2jascpal v0.2 (December 2017)
 * Copyright (C) 2017 Norbert de Jonge <mail@norbertdejonge.nl>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see [ www.gnu.org/licenses/ ].
 *
 * To properly read this code, set your program's tab stop to: 2.
 */

/*========== Includes ==========*/
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <ctype.h>
/*========== Includes ==========*/

/*========== Defines ==========*/
#define EXIT_NORMAL 0
#define EXIT_ERROR -1
#define PROG_NAME "gpl2jascpal"
#define PROG_VERSION "0.2"
#define COPYRIGHT "Copyright (C) 2017 Norbert de Jonge"
#define MAX_FILE 200
#define MAX_LINE 200
#define GIMP_PALETTE "GIMP Palette"
#define MAX_COLORS 256
#define MAX_NUMBER 10
#define MAX_WRITE 100

#ifndef O_BINARY
#define O_BINARY 0
#endif
/*========== Defines ==========*/

int iColorsRGB[3 + 2][MAX_COLORS + 2];

char *substr (char *sString, int iStart, int iChars);
int ReadGPL (char *sInFile);
int StoreLine (char *sLine, int iLine);
int WriteJASCPAL (char *sOutFile, int iRead);
int Write (char *sOutFile, char *sWrite, int iFd);
void ShowUsage (void);

/*****************************************************************************/
int main (int argc, char *argv[])
/*****************************************************************************/
{
	int iErrors;
	int iWarnings;
	int iSuccess;
	int iTemp;
	int iRead;
	int iWrite;
	char sInFile[MAX_FILE + 2];
	char sOutFile[MAX_FILE + 2];

	iErrors = 0;
	iWarnings = 0;
	iSuccess = 0;

	if (argc > 1)
	{
		for (iTemp = 1; iTemp < argc; iTemp++)
		{
			snprintf (sInFile, MAX_FILE, "%s", argv[iTemp]);
			if ((strcmp (argv[iTemp], "-h") == 0) ||
				(strcmp (argv[iTemp], "-?") == 0) ||
				(strcmp (argv[iTemp], "--help") == 0))
			{
				ShowUsage();
			} else if ((strcmp (argv[iTemp], "-v") == 0) ||
				(strcmp (argv[iTemp], "--version") == 0))
			{
				printf ("%s %s\n", PROG_NAME, PROG_VERSION);
				exit (EXIT_NORMAL);
			} else if ((sInFile[strlen (sInFile) - 4] != '.') ||
				(tolower (sInFile[strlen (sInFile) - 3]) != 'g') ||
				(tolower (sInFile[strlen (sInFile) - 2]) != 'p') ||
				(tolower (sInFile[strlen (sInFile) - 1]) != 'l'))
			{
				printf ("[FAILED] File %s doesn't have the .gpl file extension!\n",
					sInFile);
				iErrors++;
			} else if (access (sInFile, R_OK) == -1) {
				printf ("[FAILED] Could not read %s: %s!\n",
					sInFile, strerror (errno));
				iErrors++;
			} else {
				iRead = ReadGPL (sInFile);
				if (iRead == -1) { iErrors++; } else {
					if ((iRead != 16) && (iRead != 256))
					{
						printf ("[ WARN ] File %s has %i instead of 16 (4 bit) or"
							" 256 (8 bit) colors!\n", sInFile, iRead);
						iWarnings++;
					}
					snprintf (sOutFile, MAX_FILE, "%s.pal",
						substr (sInFile, 0, strlen (sInFile) - 4));
					iWrite = WriteJASCPAL (sOutFile, iRead);
					if (iWrite == -1) { iErrors++; } else {
						printf ("[  OK  ] Successfully created %s.\n", sOutFile);
						iSuccess++;
					}
				}
			}
		}
		printf ("[ INFO ] Successful conversions: %i (%i errors, %i warnings)\n",
			iSuccess, iErrors, iWarnings);
	} else {
		ShowUsage();
	}

	return EXIT_NORMAL;
}
/*****************************************************************************/
char *substr (char *sString, int iStart, int iChars)
/*****************************************************************************/
{
	char *sReturn = malloc (iChars + 1);
	strncpy (sReturn, sString + iStart, iChars);
	sReturn[iChars] = '\0';
	return sReturn;
}
/*****************************************************************************/
int ReadGPL (char *sInFile)
/*****************************************************************************/
{
	int iFd;
	int iRead;
	int iLines;
	char sChar[1 + 2];
	char sLine[MAX_LINE + 2];
	char sTemp[MAX_LINE + 2];

	iLines = -1;
	snprintf (sLine, MAX_LINE, "%s", "");

	iFd = open (sInFile, O_RDONLY);
	do {
		iRead = read (iFd, sChar, 1);
		if (iRead != 0)
		{
			if (sChar[0] != '\n')
			{
				snprintf (sTemp, MAX_LINE, "%s", sLine);
				snprintf (sLine, MAX_LINE, "%s%c", sTemp, sChar[0]);
			} else {
				if (iLines == -1)
				{
					if (strcmp (sLine, GIMP_PALETTE) != 0)
					{
						printf ("[FAILED] File %s is not a GIMP gpl palette!\n", sInFile);
						return (EXIT_ERROR);
					}
					iLines = 0;
				} else {
					if (strcmp (substr (sLine, 0, 6), "Name: ") == 0)
					{
						/*** Name ***/
					} else if (strcmp (substr (sLine, 0, 9), "Columns: ") == 0) {
						/*** Columns ***/
					} else if (sLine[0] == '#') {
						/*** Comment ***/
					} else {
						iLines++;
						if (iLines <= MAX_COLORS)
						{
							if (StoreLine (sLine, iLines) == -1) { return (EXIT_ERROR); }
						}
					}
				}
				snprintf (sLine, MAX_LINE, "%s", "");
			}
		}
	} while (iRead != 0);

	return (iLines);
}
/*****************************************************************************/
int StoreLine (char *sLine, int iLine)
/*****************************************************************************/
{
	int iTemp;
	int iStage;
	int iNumber;
	char sNumber[MAX_NUMBER + 2];
	char sTemp[MAX_NUMBER + 2];

	iStage = 0;

	for (iTemp = 0; iTemp <= (int)strlen (sLine); iTemp++)
	{
		if (isdigit (sLine[iTemp]))
		{
			if ((iStage == 0) || (iStage == 2) || (iStage == 4))
			{
				iStage++;
				snprintf (sNumber, MAX_NUMBER, "%s", "");
			}
			if ((iStage == 1) || (iStage == 3) || (iStage == 5))
			{
				snprintf (sTemp, MAX_NUMBER, "%s%c", sNumber, sLine[iTemp]);
				snprintf (sNumber, MAX_NUMBER, "%s", sTemp);
			}
		} else {
			if ((iStage == 1) || (iStage == 3) || (iStage == 5))
			{
				iNumber = atoi (sNumber);
				if ((iNumber < 0) || (iNumber > 255))
				{
					printf ("[FAILED] Incorrect color: %i!\n", iNumber);
					return (EXIT_ERROR);
				}
				switch (iStage)
				{
					case 1: iColorsRGB[0][iLine] = iNumber; break;
					case 3: iColorsRGB[1][iLine] = iNumber; break;
					case 5: iColorsRGB[2][iLine] = iNumber; break;
				}
				iStage++;
			}
		}
	}
	return (EXIT_NORMAL);
}
/*****************************************************************************/
int WriteJASCPAL (char *sOutFile, int iRead)
/*****************************************************************************/
{
	int iFd;
	int iTemp;
	int iColors;
	char sWrite[MAX_WRITE];

	if (iRead <= 16) { iColors = 16; } else { iColors = 256; }

	iFd = open (sOutFile, O_WRONLY|O_CREAT|O_EXCL|O_BINARY, 0600);
	if (iFd == -1)
	{
		printf ("[FAILED] Could not create %s: %s!\n", sOutFile, strerror (errno));
		return (EXIT_ERROR);
	} else {
		snprintf (sWrite, MAX_WRITE, "%s\r\n", "JASC-PAL");
		Write (sOutFile, sWrite, iFd);
		snprintf (sWrite, MAX_WRITE, "%s\r\n", "0100");
		Write (sOutFile, sWrite, iFd);
		snprintf (sWrite, MAX_WRITE, "%i\r\n", iColors);
		Write (sOutFile, sWrite, iFd);
		for (iTemp = 1; iTemp <= iColors; iTemp++)
		{
			if (iTemp <= iRead)
			{
				snprintf (sWrite, MAX_WRITE, "%i %i %i\r\n", iColorsRGB[0][iTemp],
					iColorsRGB[1][iTemp], iColorsRGB[2][iTemp]);
			} else {
				snprintf (sWrite, MAX_WRITE, "%s\r\n", "0 0 0");
			}
			Write (sOutFile, sWrite, iFd);
		}
	}
	return (EXIT_NORMAL);
}
/*****************************************************************************/
int Write (char *sOutFile, char *sWrite, int iFd)
/*****************************************************************************/
{
	int iWrite;

	iWrite = write (iFd, sWrite, strlen (sWrite));
	if (iWrite == -1)
	{
		printf ("[FAILED] Could not write to %s: %s!\n",
			sOutFile, strerror (errno));
		return (EXIT_ERROR);
	} else {
		return (EXIT_NORMAL);
	}
}
/*****************************************************************************/
void ShowUsage (void)
/*****************************************************************************/
{
	printf ("%s %s\n%s\n\n", PROG_NAME, PROG_VERSION, COPYRIGHT);
	printf ("Usage:\n");
	printf ("  %s [OPTIONS] <file(s).gpl>\n\nOptions:\n", PROG_NAME);
	printf ("  -h, -?,    --help           display this help and exit\n");
	printf ("  -v,        --version        output version information and"
		" exit\n");
	exit (EXIT_NORMAL);
}
/*****************************************************************************/
